from datetime import datetime

from flask import request

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.date_interval import DateInterval
from flask_monitoringdashboard.core.reporting.questions.average_latency import AverageLatency
from flask_monitoringdashboard.core.reporting.questions.status_code_distribution import StatusCodeDistribution
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_endpoints


def get_date(p):
    return datetime.utcfromtimestamp(int(request.args.get(p)))


def make_endpoint_summary(endpoint, comparison_interval, compared_to_interval):
    questions = [AverageLatency(), StatusCodeDistribution()]

    summary = dict(
        endpoint_id=endpoint.id,
        endpoint_name=endpoint.name,
        answers=[],
        has_anything_significant=False
    )

    for question in questions:
        answer = question.get_answer(endpoint, comparison_interval, compared_to_interval)

        if answer.is_significant():
            summary['has_anything_significant'] = True

        summary['answers'].append(answer.serialize())

    return summary


@blueprint.route('/api/reporting/make_report', methods=['POST'])
@secure
def make_report():
    arguments = request.json

    try:
        comparison_interval = DateInterval(datetime.fromtimestamp(int(arguments['comparison_interval']['from'])),
                                           datetime.fromtimestamp(int(arguments['comparison_interval']['to'])))

        compared_to_interval = DateInterval(datetime.fromtimestamp(int(arguments['compared_to_interval']['from'])),
                                            datetime.fromtimestamp(int(arguments['compared_to_interval']['to'])))

    except Exception:
        return 'Invalid payload', 422

    endpoint_summaries = []
    with session_scope() as db_session:
        for endpoint in get_endpoints(db_session):
            endpoint_summary = make_endpoint_summary(endpoint, comparison_interval, compared_to_interval)
            endpoint_summaries.append(endpoint_summary)

    return dict(
        summaries=endpoint_summaries
    )



from sqlalchemy import desc
from sqlalchemy.orm import joinedload

from flask_monitoringdashboard.database import Outlier, Request


def add_outlier(db_session, request_id, cpu_percent, memory, stacktrace, request):
    """
    Adds an Outlier object in the database.
    :param db_session: session for the database
    :param request_id: id of the request
    :param cpu_percent: cpu load of the server when processing the request
    :param memory: memory load of the server when processing the request
    :param stacktrace: stack trace of the request
    :param request: triple containing the headers, environment and url
    """
    headers, environ, url = request
    outlier = Outlier(
        request_id=request_id,
        request_header=headers,
        request_environment=environ,
        request_url=url,
        cpu_percent=cpu_percent,
        memory=memory,
        stacktrace=stacktrace,
    )
    db_session.add(outlier)


def get_outliers_sorted(db_session, endpoint_id, offset, per_page):
    """
    Gets a list of Outlier objects for a certain endpoint, sorted by most recent request time
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint for filtering the requests
    :param offset: number of items to skip
    :param per_page: number of items to return
    :return list of Outlier objects of a specific endpoint
    """
    result = (
        db_session.query(Outlier)
        .join(Outlier.request)
        .options(joinedload(Outlier.request).joinedload(Request.endpoint))
        .filter(Request.endpoint_id == endpoint_id)
        .order_by(desc(Request.time_requested))
        .offset(offset)
        .limit(per_page)
        .all()
    )
    db_session.expunge_all()
    return result


def get_outliers_cpus(db_session, endpoint_id):
    """
    Gets list of CPU loads of all outliers of a certain endpoint
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return list of cpu percentages as strings
    """
    outliers = (
        db_session.query(Outlier.cpu_percent)
        .join(Outlier.request)
        .filter(Request.endpoint_id == endpoint_id)
        .all()
    )
    return [outlier[0] for outlier in outliers]



from __future__ import division
import numpy as np


def mean_diff(x, y):
    return np.abs(np.mean(x) - np.mean(y))


def mean_permutation_test(x, y, num_rounds=1000):
    """
    Performs a non-parametric test to check whether `x` and `y` come from the same distribution.

    :param x: a sample from some distribution
    :param y: a sample to compare x to
    :param num_rounds: number of different permutations to test. Increase this number to increase
     the accuracy
    :return: The p-value
    """
    rng = np.random.RandomState()

    m = len(x)
    combined = np.hstack((x, y))

    more_extreme = 0
    reference_stat = mean_diff(x, y)

    for i in range(num_rounds):
        rng.shuffle(combined)

        if mean_diff(combined[:m], combined[m:]) > reference_stat:
            more_extreme += 1

    return more_extreme / num_rounds



import unittest

from flask_monitoringdashboard.test.utils import (
    set_test_environment,
    clear_db,
    add_fake_data,
    get_test_app,
    test_admin_secure,
    NAME,
)


class TestTestMonitor(unittest.TestCase):
    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_build_performance(self):
        test_admin_secure(self, 'test_build_performance')

    def test_endpoint_build_performance(self):
        test_admin_secure(self, 'endpoint_build_performance')

    def test_testmonitor(self):
        test_admin_secure(self, 'testmonitor')

    def test_endpoint_test_details(self):
        test_admin_secure(self, 'testmonitor/{}'.format(NAME))
